(* Copyright (c) 2011 - 2015. Cary Jensen, Jensen Data Systems, Inc.
   This code sample is intended for the readers of
   "Delphi in Depth: ClientDataSets"
     and
   "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen

   No guarantees or warranties are expressed or implied concerning
   the applicability of techniques or code included in this example
   or in the accompanying book. If you wish to use techniques or
   code included in this example or described in the book, it is
   your responsibility to test and certify any code, techniques,
   or design you adopt.

   For information on consulting or training services, please visit:
   http://www.JensenDataSystems.com

   For more information about "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen, including links to order the book, please visit:
   http://www.JensenDataSystems.com/cdsbook2
*)
unit mainformu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DBTables, Db, StdCtrls, DBCtrls, Grids, DBGrids, ExtCtrls, DBClient,
  Provider;

type
  TForm1 = class(TForm)
    DataSource1: TDataSource;
    Query1: TQuery;
    Panel2: TPanel;
    DBGrid1: TDBGrid;
    Label1: TLabel;
    Splitter1: TSplitter;
    Panel3: TPanel;
    Label2: TLabel;
    DBGrid2: TDBGrid;
    Table1: TTable;
    DataSource2: TDataSource;
    DataSetProvider1: TDataSetProvider;
    ClientDataSet1: TClientDataSet;
    Panel1: TPanel;
    DBNavigator1: TDBNavigator;
    Button2: TButton;
    Query2: TQuery;
    procedure FormCreate(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure DataSetProvider1BeforeUpdateRecord(Sender: TObject;
      SourceDS: TDataSet; DeltaDS: TCustomClientDataSet;
      UpdateKind: TUpdateKind; var Applied: Boolean);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
var
  Qry: TQuery;
  Tbl: TTable;
begin
Tbl := TTable.Create(Self);
try
  Qry := TQuery.Create(Self);
  try
    Qry.DatabaseName := 'DBDEMOS';
    Tbl.DatabaseName := ExtractFilePath(Application.ExeName);
    Tbl.TableName := 'DemoTabl.db';
    //Define the SQL statement
    Qry.Sql.Add('SELECT	CUSTOMER."Company" , ORDERS."OrderNo" ,');
    Qry.Sql.Add('	ITEMS."PartNo", ORDERS."EmpNo"');
    Qry.Sql.Add('FROM	"CUSTOMER.DB" CUSTOMER , "ORDERS.DB" ORDERS ,');
    Qry.Sql.Add('	"ITEMS.DB" ITEMS');
    Qry.Sql.Add('WHERE	( CUSTOMER.CustNo = ORDERS.CustNo )');
    Qry.Sql.Add('		AND');
    Qry.Sql.Add('	( ORDERS.OrderNo = ITEMS.OrderNo )');
    Qry.Open;
    //Copy the records to a Table
    Tbl.BatchMove(Qry,batCopy);
    //Add a primary and a secondary index
    Tbl.AddIndex('firstidx','Company;OrderNo;PartNo',[ixPrimary, ixUnique]);
    Tbl.AddIndex('partidx','PartNo',[ixCaseInsensitive]);
    Qry.Close;
    //Execute a query against the newly created table.
    Query1.DatabaseName := Tbl.DatabaseName;
    Query1.Open;
    Table1.DatabaseName := Tbl.DatabaseName;
    Table1.TableName := Tbl.TableName;
  finally
    Tbl.Free;
  end;
  Table1.Open;
  ClientDataSet1.Open;
  Query2.Prepare;
finally
  Qry.Free;
end;

end;

procedure TForm1.Button2Click(Sender: TObject);
begin
if ClientDataSet1.State in [dsInsert, dsEdit] then
  ClientDataSet1.Post;
if ClientDataSet1.ChangeCount = 0 then
  begin
    ShowMessage('No changes to apply');
    Exit;
  end;
ClientDataSet1.ApplyUpdates(-1);
Table1.Refresh;
end;

procedure TForm1.DataSetProvider1BeforeUpdateRecord(Sender: TObject;
  SourceDS: TDataSet; DeltaDS: TCustomClientDataSet;
  UpdateKind: TUpdateKind; var Applied: Boolean);
begin
  Applied := True;
  if UpdateKind = ukModify then
  begin
    Query2.ParamByName('PartNo').Value :=
      DeltaDS.FieldByName('PartNo').NewValue;
    Query2.ParamByName('OLD_PartNo').Value :=
      DeltaDS.FieldByName('PartNo').OldValue;
    Query2.ExecSQL;
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
Query2.UnPrepare;
end;

end.
