(* Copyright (c) 2011 - 2015. Cary Jensen, Jensen Data Systems, Inc.
   This code sample is intended for the readers of
   "Delphi in Depth: ClientDataSets"
     and
   "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen

   No guarantees or warranties are expressed or implied concerning
   the applicability of techniques or code included in this example
   or in the accompanying book. If you wish to use techniques or
   code included in this example or described in the book, it is
   your responsibility to test and certify any code, techniques,
   or design you adopt.

   For information on consulting or training services, please visit:
   http://www.JensenDataSystems.com

   For more information about "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen, including links to order the book, please visit:
   http://www.JensenDataSystems.com/cdsbook2
*)
unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, Provider, StdCtrls, ExtCtrls, DBClient, DB, DBTables,
  Grids, DBGrids, DBCtrls, mmsystem;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    DBNavigator1: TDBNavigator;
    DBGrid1: TDBGrid;
    DataSource1: TDataSource;
    Panel2: TPanel;
    DBNavigator2: TDBNavigator;
    DBGrid2: TDBGrid;
    DataSource2: TDataSource;
    Table1: TTable;
    ClientDataSet1: TClientDataSet;
    DataSetProvider1: TDataSetProvider;
    Splitter1: TSplitter;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    SelectTable1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    View1: TMenuItem;
    Delta1: TMenuItem;
    TableLoading1: TMenuItem;
    OpenDialog1: TOpenDialog;
    ApplyUpdates1: TMenuItem;
    ClientDataSet2: TMenuItem;
    Load1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    CancelUpdates1: TMenuItem;
    GetNextPacket1: TMenuItem;
    RuntimeConfiguration1: TMenuItem;
    UseOnReconcileErrorDialogBox1: TMenuItem;
    UseOnUpdateErrorEventHandler1: TMenuItem;
    ResolvetoDataSet1: TMenuItem;
    procedure ApplyUpdates1Click(Sender: TObject);
    procedure TableLoading1Click(Sender: TObject);
    procedure SelectTable1Click(Sender: TObject);
    procedure Load1Click(Sender: TObject);
    procedure Delta1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure CancelUpdates1Click(Sender: TObject);
    procedure GetNextPacket1Click(Sender: TObject);
    procedure ClientDataSet1AfterOpen(DataSet: TDataSet);
    procedure ClientDataSet1AfterClose(DataSet: TDataSet);
    procedure RuntimeConfiguration1Click(Sender: TObject);
    procedure Table1AfterInsert(DataSet: TDataSet);
    procedure Table1AfterPost(DataSet: TDataSet);
    procedure Table1AfterDelete(DataSet: TDataSet);
    procedure UseOnReconcileErrorDialogBox1Click(Sender: TObject);
    procedure UseOnReconcileErrorEventHandler1Click(Sender: TObject);
    procedure ClientDataSet1ReconcileError(DataSet: TCustomClientDataSet;
      E: EReconcileError; UpdateKind: TUpdateKind;
      var Action: TReconcileAction);
    procedure ResolvetoDataSet1Click(Sender: TObject);
    procedure DataSetProvider1UpdateError(Sender: TObject;
      DataSet: TCustomClientDataSet; E: EUpdateError; UpdateKind: TUpdateKind;
      var Response: TResolverResponse);
    procedure DataSetProvider1Validate(const Delta: OleVariant);
    procedure Exit1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses deltaform, dynamicform, RecError;

{$R *.dfm}

procedure TForm1.ApplyUpdates1Click(Sender: TObject);
begin
  ClientDataSet1.ApplyUpdates(2);
end;

procedure TForm1.ClientDataSet1AfterClose(DataSet: TDataSet);
begin
  GetNextPacket1.Visible := False;
end;

procedure TForm1.ClientDataSet1AfterOpen(DataSet: TDataSet);
begin
  GetNextPacket1.Visible := not ClientDataSet1.FetchOnDemand;
end;

procedure TForm1.ClientDataSet1ReconcileError(DataSet: TCustomClientDataSet;
  E: EReconcileError; UpdateKind: TUpdateKind; var Action: TReconcileAction);
begin
  if UseOnReconcileErrorDialogBox1.Checked then
    Action := HandleReconcileError(DataSet, UpdateKind, E)
  else
    ShowMessage('Error : ' + e.Message);
end;

procedure TForm1.GetNextPacket1Click(Sender: TObject);
begin
if ClientDataSet1.GetNextPacket < ClientDataSet1.PacketRecords then
  ShowMessage('There are no more records to fetch');
end;

procedure TForm1.Table1AfterDelete(DataSet: TDataSet);
begin
  ShowMessage('A record has been deleted');
end;

procedure TForm1.Table1AfterInsert(DataSet: TDataSet);
begin
  ShowMessage('A record has been inserted');
end;

procedure TForm1.Table1AfterPost(DataSet: TDataSet);
begin
  ShowMessage('A record has been updated');
end;

procedure TForm1.TableLoading1Click(Sender: TObject);
begin
  if TableLoading1.Checked then
  begin
    Table1.DisableControls;
    TableLoading1.Checked := False;
  end
  else
  begin
    Table1.EnableControls;
    TableLoading1.Checked := True;
  end;
end;

procedure TForm1.UseOnReconcileErrorDialogBox1Click(Sender: TObject);
begin
  UseOnReconcileErrorDialogbox1.Checked :=
    not UseOnReconcileErrorDialogbox1.Checked;
end;

procedure TForm1.UseOnReconcileErrorEventHandler1Click(Sender: TObject);
begin
  if UseOnUpdateErrorEventHandler1.Checked then
  begin
    UseOnUpdateErrorEventHandler1.Checked := False;
    ClientDataSet1.OnReconcileError := nil;
  end
  else
  begin
    UseOnUpdateErrorEventHandler1.Checked := True;
    ClientDataSet1.OnReconcileError := ClientDataSet1ReconcileError;
  end;
end;

procedure TForm1.SelectTable1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    if Table1.Active then
      Table1.Close;
    Table1.TableName := OpenDialog1.FileName;
  end;
end;

procedure TForm1.Load1Click(Sender: TObject);
begin
if ClientDataSet1.Active then
begin
  Load1.Caption := 'Load';
  ClientDataSet1.Close;
end
else
begin
  Load1.Caption := 'Close';
  ClientDataSet1.Open;
end;
end;

procedure TForm1.ResolvetoDataSet1Click(Sender: TObject);
begin
  if ResolvetoDataSet1.Checked then
  begin
    ResolvetoDataSet1.Checked := False;
    DataSetProvider1.ResolveToDataSet := False;
  end
  else
  begin
    ResolvetoDataSet1.Checked := True;
    DataSetProvider1.ResolveToDataSet := True;
  end;
end;

procedure TForm1.RuntimeConfiguration1Click(Sender: TObject);
begin
  with TRuntimeForm.Create(nil) do
    Show;
end;

procedure TForm1.DataSetProvider1UpdateError(Sender: TObject;
  DataSet: TCustomClientDataSet; E: EUpdateError; UpdateKind: TUpdateKind;
  var Response: TResolverResponse);
begin
  ShowMessage('OnUpdateError');
end;

procedure TForm1.DataSetProvider1Validate(const Delta: OleVariant);
begin
 ShowMessage('OnValidate');

end;

procedure TForm1.Delta1Click(Sender: TObject);
begin
if not Self.ClientDataSet1.Active then
  begin
    ShowMessage('ClientDataSet is not active');
    Exit;
  end;
if Self.ClientDataSet1.ChangeCount = 0 then
  begin
    ShowMessage('Delta is empty');
    Exit;
  end;

  with TChangeCache.Create(nil) do
  begin
    ClientDataSet1.Data := Self.ClientDataSet1.Delta;
    ShowModal;
    Release;
  end;
end;

procedure TForm1.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TForm1.About1Click(Sender: TObject);
begin
ShowMessage('CDSLoadBehaviorDemo (c) 2002 Cary Jensen, Jensen Data Systems, Inc. '
  + #10#13 + ''
  + #10#13 + 'Demonstration of the behavior of ClientDataSets and DataSetProviders.'
  + #10#13 + 'This project is provided for demonstration purposes only.'
  + #10#13 + ''
  + #10#13 + 'Visit http://www.JensenDataSystems.com for information about training and consulting services.');

end;

procedure TForm1.CancelUpdates1Click(Sender: TObject);
begin
ClientDataSet1.CancelUpdates;
end;

end.
